import os
import json 
import hashlib
import requests
import gitlab
from atlassian import Confluence
from markdownify import markdownify as md
import requests, json
import urllib3
from sagemaker.jumpstart.model import JumpStartModel, AsyncInferenceConfig
from sagemaker.serializers import JSONSerializer 
from sagemaker.deserializers import JSONDeserializer
from langchain_community.embeddings import SagemakerEndpointEmbeddings
from langchain_community.embeddings.sagemaker_endpoint import EmbeddingsContentHandler
from langchain.chains import LLMChain
from langchain_community.llms import SagemakerEndpoint 
from langchain_core.prompts import PromptTemplate
from langchain_community.llms.sagemaker_endpoint import LLMContentHandler
from langchain_core.documents import Document
import boto3
import base64
from langchain_openai import AzureChatOpenAI
import httpx
import logging
from typing import Optional

#will use IAM roles instead of access keys

aws_access_key_id = os.environ.get("AWS_ACCESS_KEY_ID") 
aws_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY")
aws_session_token = os.environ.get("AWS_SESSION_TOKEN")
project_id = int(os.environ.get("CI_PROJECT_ID"))
author = str(os.environ.get("CI_COMMIT_AUTHOR"))
gitlab_token = os.environ.get("CI_JOB_TOKEN")
job_id=int(os.environ.get("CI_JOB_ID"))
confluence_token = str(os.getenv("CONFLUENCE_TOKEN"))
confluence_username = str(os.getenv("CONFLUENCE_USERNAME"))
confluence_id = int(os.getenv("CONFLUENCE_DOC_ID"))
ignored_files = os.getenv("IGNORED_FILES")
system_prompt = str(os.getenv("SYSTEM_PROMPT"))
default_model = str(os.environ.get("DEFAULT_MODEL"))
pass_fail_config = json.loads(os.environ.get("PASSFAIL_CONFIG"))
pipeline_source = os.environ.get("CI_PIPELINE_SOURCE")
pipeline_id = int(os.environ.get("CI_PIPELINE_ID"))
curr_branch = str(os.environ.get("CI_COMMIT_BRANCH"))
#TODO: handle undefined variables

#------LOGGER SETTINGS---------------------------------------------#

logging.getLogger('boto3').setLevel(logging.INFO)
logging.getLogger('botocore').setLevel(logging.INFO)
logging.getLogger('urllib3').setLevel(logging.INFO)

logger = logging.getLogger(__name__)
with open('auto-code-review-script-logs.txt', "w") as f:
    logging.basicConfig(filename='auto-code-review-script-logs.txt', encoding='utf-8', level=logging.DEBUG)

#------LLM SECTION----------------------------------------------#

# Reading in Confluence guideline page
if(confluence_username is not None and confluence_token is not None):
    confluence = Confluence(
        url='https://ge-hc.atlassian.net/wiki/',
        username = confluence_username,
        password = confluence_token,
        verify_ssl = False
    )
    if(confluence_id is not None):
        try:
            guideline_doc = confluence.get_page_by_id(confluence_id, expand='body.storage')
        except:
            logger.error("Invalid Confluence Page ID")
else:
    logger.info("No Confluence information provided")
    guideline_doc = """Use consistent naming conventions: Consistent naming conventions should be used throughout the code. For example, use CamelCase for method names and variables, and PascalCase for classes.
Use exception handling: Exception handling should be used to handle errors and exceptions that may occur during runtime.
Use proper indentation and formatting: Proper indentation and formatting should be used to make the code easy to read and understand.
Use version control: Version control should be used to track changes to the code and collaborate with other developers.
Use testing: Testing should be used to ensure that the code is working as expected and to catch any bugs or errors.
Use design patterns: Design patterns should be used to solve common problems and improve the architecture of the code.
Follow the SOLID principles: The SOLID principles (Single Responsibility, Open-Closed, Liskov Substitution, Interface Segregation, and Dependency Inversion) should be followed to create well-structured and maintainable code."""

def sagemaker_model():
    endpoint_name = 'sagemaker-inference-ep-development' 
    sagemaker_client = boto3.client(
        'runtime.sagemaker',
        region_name='us-east-1',
        aws_access_key_id=aws_access_key_id, 
        aws_secret_access_key=aws_secret_access_key,
        aws_session_token=aws_session_token, 
        verify=False
    )
    class LLama3ContentHandler(LLMContentHandler):
        content_type = "application/json"
        accepts = "application/json" 

        def transform_input(self, prompt: str, model_kwargs: dict) -> bytes:
            input_str = json.dumps({"inputs": prompt, "parameters": model_kwargs})
            return input_str.encode("utf-8")
    
        def transform_output(self, output: 'boto3.streamingbody.StreamingBody') -> str:
            response_str = output.read().decode("utf-8")
            response_json = json.loads(response_str)
            list = response_json[0]["generated_text"].split("assistant") 
            assistantResponse = list[len(list) - 1].split("assistant")
            return assistantResponse[len(assistantResponse) - 1].strip("").strip("\\n\\n") 
    content_handler = LLama3ContentHandler()
    model_kwargs = {
        "max_length": 300,
        "num_return_sequences": 1, 
        "do_sample": True,
        #choosing the words that are 90% likely to come next
        "top_p": 0.9, 
        "temperature": 0.85,
        "max_new_tokens": 512,
        "stop": ["<|endoftext|>", "</s>"], 
        #no repeats 
        "early_stopping": True
    }
    #invoking Llama3 Sagemaker endpoint 
    sagemaker_llm = SagemakerEndpoint(
            endpoint_name=endpoint_name,
            model_kwargs=model_kwargs,         
            content_handler=content_handler,
            client=sagemaker_client,
            endpoint_kwargs={"TargetVariant": "Meta-Llama-3-8B-Instruct-variant"}
    )

    return get_chain(sagemaker_llm)

def openai_model(default_model):
    OPENAI_ENDPOINT = "https://openai-llm-frontdoor-hma7evbthrd4cugn.a01.azurefd.net/"
    if(default_model == "gpt-4"):
        OPENAI_DEPLOYMENT_MODEL = "gpt-4-32k-beta"
    else:
        OPENAI_DEPLOYMENT_MODEL = "gpt-4o"
    OPENAI_AZURE_API_VERSION = "2023-12-01-preview"

    def get_llm_access_token() -> str:
            IDAM_TOKEN_URL =  "https://idam.gehealthcloud.io/oauth2/token" 
            LLM_APP_AUDIENCE = "0_b2dJB20TBhxzLIHCMzSG4RiQYa"
            APP_CLIENT_ID = "genai-dev-chat-bot-client-id"
            APP_CLIENT_SECRET = "genai-dev-chat-bot-client-secret"
            """
            This function generate the token with onco app credentials and exchange the token with LLMaS app.
            :return: Access Token
            """
            payload = f'grant_type=client_credentials&client_id={APP_CLIENT_ID}&client_secret={APP_CLIENT_SECRET}'
            headers = {
                'Content-Type': 'application/x-www-form-urlencoded'
            }
            response = requests.request("POST", IDAM_TOKEN_URL, headers=headers, data=payload, verify=False)
            token = response.json()['access_token']
            # Token exchange.
            payload = f'grant_type=token_exchange&token={token}&audience={LLM_APP_AUDIENCE}'
            basic_auth_creds = f"{APP_CLIENT_ID}:{APP_CLIENT_SECRET}"
            basic_auth_creds = base64.b64encode(basic_auth_creds.encode()).decode("ascii")
            headers = {
                'Content-Type': 'application/x-www-form-urlencoded',
                'Authorization': f'Basic {basic_auth_creds}'
            }
            response = requests.request("POST", IDAM_TOKEN_URL, headers=headers, data=payload, verify=False)
            return response.json()['access_token']

    gpt_llm = AzureChatOpenAI(
            api_key="xxx",  #Not playing any role, but required as per OpenAI sdk. So any random could be passed.
            azure_endpoint=OPENAI_ENDPOINT,
            azure_deployment=OPENAI_DEPLOYMENT_MODEL,
            api_version=OPENAI_AZURE_API_VERSION,
            http_client=httpx.Client(verify=False),
            default_headers={
                'Authorization': f'Bearer {get_llm_access_token()}',
                'Content-Type': 'application/json'
            },
            temperature=0.8
    )

    return get_chain(gpt_llm)

def get_chain(llm):
    prompt_template = """ {system_prompt}

    #GUIDELINES#: {guidelines}

    Now, write a review for the following piece of code:  \n {code} 

    \nBegin Review:\n"""
    
    PROMPT = PromptTemplate(
        input_variables=["system_prompt","code", "guidelines"], template=prompt_template 
    )

    llm_chain = LLMChain(llm=llm, prompt=PROMPT)
    return llm_chain
 
def get_default_model():
    return default_model

#------GITLAB SECTION----------------------------------------------#
gl = gitlab.Gitlab(url='https://gitlab.apps.ge-healthcare.net', private_token=gitlab_token)
project = gl.projects.get(project_id)

# turns notifications on automatically for project
settings = gl.projects.get(project_id).notificationsettings.get()
settings.new_issue = True
settings.new_note = True
settings.save()

# turns output string into json format
def clean_output(s):
    s.replace('\'', '\"') 
    s = s[s.find("["):(s.find("]")+1)]
    if len(s) == 0: 
        return s
    else:
        try:
            return json.loads(s) 
        except json.JSONDecodeError as e:
            try:
                json.loads(s[:e.pos])   
            except:
                return ""

def is_valid_format(res,length):
    if(res==None or len(res) <= 1):
        return False
    for j in res:
        try:
            i = j["Line Number"]
        except:
            return False
            try:
                int_j = int(j["Line Number"])
                if(int_j > length):
                    return False
            except:
                return False
    return True 

# contains prompt and langchain definitions, returns json-formatted feedback from llm
def get_llm_output(llm_chain, filename):
    with open(filename, "r") as f:
        file_content = f.read()
    try:
        output = llm_chain.invoke(input={"system_prompt":system_prompt, "code":file_content, "guidelines":guideline_doc})
    except: # if there is an authentication issue with gpt4, use llama3 to get output instead
        logger.info("Issue authenticating with GPT, using Llama-3 instead.")
        default_model = "llama-3"
        return get_llm_output(sagemaker_model(), filename)
    comments = output['text']
    comments = clean_output(comments)    
    # ensures that res is in the correct format before it gets returned. If not in correct format, the the chain is re-run
    if(not is_valid_format(comments,file_content.count('\n'))):
        logger.info("Invalid format, re-running chain")
        return get_llm_output(llm_chain, filename)
    return comments

#takes LLM comments and prints them on commit diff
def print_comments_on_commit(res, filename, commit):
    with open(filename, "r") as f:
        file_content = f.read()
    for j in res:
        if(("-") in j["Line Number"]):
            int_j = j["Line Number"].split("-")[1]
        else:
            if(j["Line Number"]=="NA"): #if it is a general comment, post as a "note" on Gitlab MR
                int_j = 1
            else:
                int_j = int(j["Line Number"])
        try:
            if(j["Code Suggestion"] is not None and len(j["Code Suggestion"])>0):
                note = "On line "+ str(int_j) + " of " + filename +": "+ j['Comment'] + "\nSuggestion: " + j['Code Suggestion']
            else:
                note = "On line "+ str(int_j) + " of " + filename +": "+ j['Comment']
        except:
            note = "On line "+ str(int_j) + " of " + filename +": "+ j['Comment']
        logger.info(note)
        commit.comments.create({'note': note,'line': int_j,'sha':commit.refs('branch'),'line_type': 'new', 'path': filename})
    return res

#takes LLM comments and prints them on mr diff
def print_comments_on_mr(res, filename, mr):
        diffs = mr.diffs.list(get_all=False)
        mr_diff = diffs[0]
        with open(filename, "r") as f:
            file_content = f.read()
        for j in res:
            try:
                if(j["Code Suggestion"] != ""): #if there is a code suggestion, format as suggestion and include comment after
                    note_content = "```suggestion:-0+0\n" + j["Code Suggestion"] + "\n```\n" + j["Comment"]
                else:
                    note_content = "On line(s) " + j["Line Number"] + " of " +str(filename) + ": " + j["Comment"]
            except:
                note_content = "On line(s) " + j["Line Number"] + " of " +str(filename) + ": " + j["Comment"]
            if(j["Line Number"]=="NA" or "-" in j["Line Number"] or "," in j["Line Number"]): #if it is a general comment, post as a "note" on Gitlab MR
                mr.notes.create({'body': note_content})
            else:
                try:
                    int_j = int(j["Line Number"])
                    try:
                        mr.discussions.create({'body': note_content,
                                                    'position': {
                                                        "base_sha": mr_diff.base_commit_sha,
                                                        "head_sha": mr_diff.head_commit_sha,
                                                        "start_sha": mr_diff.start_commit_sha,
                                                        'position_type': 'text',
                                                        'new_line': int_j,
                                                        'linetype': 'new',
                                                        'new_path': filename}
                                                    })
                    except:
                            mr.notes.create({'body': note_content})
                except Exception as e:
                    print(e)
                    mr.notes.create({'body': note_content})
                        
#takes in the LLM comments and dictionary (key = severity, value = num_occurences_allowed) to determine if the pipeline should pass or fail

def pipeline_passes(res):
    counts = {
    "1": 0,
    "2": 0,
    "3": 0,
    "4": 0,
    "5": 0
    }
    for i in res:
        counts[str(i['Severity'])] += 1
    for j in range(1,6):
        if(int(counts[str(j)])>int(pass_fail_config[str(j)])):
            return False
    return True

if(default_model == "gpt-4o"):
    llm_chain = openai_model("gpt-4o")
    logger.info("Using GPT-4o")
elif (default_model == "gpt-4"):
    logger.info("Using GPT-4")
    llm_chain = openai_model("gpt-4")
else:
    llm_chain = sagemaker_model()
    logger.info("Using Llama-3 hosted on Sagemaker")


if(pipeline_source == "merge_request_event"):
    logger.info("Running merge request pipeline ")

    mr_iid = int(os.environ.get("CI_MERGE_REQUEST_IID"))
    mr= project.mergerequests.get(mr_iid)
    changes = mr.changes()['changes']
    try:
        pipeline = project.pipelines.get(pipeline_id)
        logger.info("Running on pipeline " + str(pipeline_id))
    except:
        logger.warn("Invalid pipeline id. Using most recent pipeline.")
        pipeline = project.pipelines.list(get_all=False)[0]
    user_id = pipeline.user['id']

    #--------COMMENTING SECTION-------------------------------------------#
    passed = True
    summary = "**<h3>Summary:</h3>** \n\n"

    # loops through each file in diff
    for i in changes:     
        if(i['deleted_file'] == False and (i['old_path']==i['new_path'])): #if it is just a changed file
            filename = i['new_path']
            if (filename not in ignored_files):
                logger.info("Reviewing " + filename + ": ")
                res = get_llm_output (llm_chain, filename)
                for j in res:
                    if(j['Severity']==5):
                        summary += "- [ ] :warning:"
                    elif(j['Severity']>3):
                        summary += "- [ ] :stop_sign:"
                    else:
                        summary += "- [ ] "
                    summary += (" **On line ")
                    summary += j['Line Number']
                    summary += " of " 
                    summary += filename + ":** "
                    summary += j['Comment'] 
                    summary += "\n\n"
                summary += "*These comments were generated by " + str(get_default_model()) + "*" + "\n\n"
                #updating the passed variable
                passed = passed and pipeline_passes(res)
                print_comments_on_mr(res, filename, mr)

    emojis = mr.awardemojis.list()
    if(passed):
        logger.info("Pipeline has been passed.")
        if(len(emojis) != 0):
            for emoji in emojis:
                emoji.delete()
        emoji = mr.awardemojis.create({'name': 'robot'})
        emoji = mr.awardemojis.create({'name': 'thumbsup'})
    else:
        logger.info("Pipeline has been failed.")
        issue = project.issues.create({'title': "Review Merge Request", 'description': summary})
        issue.assignee_ids = [user_id]
        issue.save()
        if(len(emojis) != 0):
            for emoji in emojis:        
                emoji.delete()
        emoji = mr.awardemojis.create({'name': 'robot'})
        emoji = mr.awardemojis.create({'name': 'thumbsdown'})
        raise Exception("Serious issues identified. Please review the comments on this MR detailing the issues and make necessary edits.")
else:
    logger.info("Running commit pipeline")
    commits = project.commits.list(ref_name=curr_branch, get_all=False)
    commit = commits[0] 
    diff = commit.diff()
    passed = True
    logger.info("Running on pipeline " + str(pipeline_id))

    # loops through each file in diff
    for i in range(0,len(diff)):
        filename = diff[i]['new_path']
        logger.info("Reviewing " + filename + ": ")
        if (filename not in ignored_files and diff[i]['deleted_file'] == False): #validating if we should look at this file
            res =  print_comments_on_commit(get_llm_output(llm_chain, filename), filename, commit)
            commit.comments.create({'note': "*These comments were generated by " + get_default_model() + "*",'line': 1,'sha':commit.refs('branch'),'line_type': 'new', 'path': filename})
            passed = pipeline_passes(res) and passed
        else:
            logger.info(str(filename) + " either deleted or ignored.")

    # fails the pipeline if the code is not passed
    if(not passed):
        logger.warn("Pipeline has been failed.")
        raise Exception("Serious issues identified. Please review the comments on this commit and make necessary edits.")
    else:
        logger.info("Pipeline has succeeded.")

